#!/bin/bash

### The postprocessing part of the neural network. This is run at each generation, after all individual chunks are
### completed.

### 1. Loops the directories of each chunk and collects the cost value. Appends this to a copy of 00.csv and saves as a
###    new file, so that there will ultimately be a .csv file of parameters and their associated costs at every
###    generation
### 2. Performs a genetic algorithm (Biased Random Key Genetic Algorithm (http://mauricio.resende.info/doc/brkgaAPI.pdf)
###    the set of parameters that are generated are saved onto 00.csv, to be used in the next generation.


### 1. Loops the directories of each chunk and collects the cost value. Appends this to a copy of 00.csv and saves as a
###    new file, so that there will ultimately be a .csv file of parameters and their associated costs at every
###    generation

# SAME AS pre.sh
MODEL=/gpfs/scratch/bsc32/bsc32786/%EXPID%/EXP00/
VARIABLES=${MODEL}variables.csv # path to the information about the variables (file must be called variables.csv) (PATH AND FILE)


module load python

function makecsv {

parameters="$(python - $1 << END
# $1 - path to directories of the chunks
# copies 0000.csv to a new file (numbered according to generation number), makes a new column then loops through each
# chunk, reads cost.txt, adds each to the column

import numpy as np
import os
import pandas

dir='$1'
os.chdir(dir)

chunks=os.listdir(os.getcwd())
chunks=[i for i in chunks if os.path.isdir(i) if len(i)==4 if i.isdigit()]

os.chdir(dir+'parameters')
parameters=np.genfromtxt('0000.csv')
parameters=np.append(parameters,np.zeros((np.shape(parameters)[0],1)),axis=1)

bbp700stats=np.zeros((parameters.shape[0],5))
chlastats=np.zeros((parameters.shape[0],5))

names=['bias','normstd','R','RMSE','cost']

for chunk in chunks:
    ind=int(chunk)-1
    os.chdir(dir+'/'+chunk)

    # appending the cost function that will be used in the GA
    #cost=np.genfromtxt('bbp700cost.csv')
    cost=np.genfromtxt('bPOCcost.csv')
    parameters[ind,-1]=cost

    # creating .csv files of other statistical parameters
    for i in range(len(names)):
        #x=np.genfromtxt('bbp700'+names[i]+'.csv')
        x=np.genfromtxt('bPOC'+names[i]+'.csv')
        bbp700stats[ind,i]=x
        x=np.genfromtxt('chla'+names[i]+'.csv')
        chlastats[ind,i]=x


bbp700stats=pandas.DataFrame(bbp700stats,columns=names)
chlastats=pandas.DataFrame(chlastats,columns=names)

os.chdir(dir+'/parameters')
gen=len(os.listdir(dir+'/parameters'))
print(gen)
gen="{0:0=4d}".format(gen)+'.csv'

np.savetxt(gen,parameters)

os.chdir(dir+'default')
bbp700refbias=np.genfromtxt('bbp700refbias.csv')

maxbias=max(abs(bbp700stats['bias']))
#maxbias=maxbias*bbp700refbias

#os.remove("bbp700refbias.csv")
#np.savetxt('bbp700refbias.csv',[maxbias])

os.remove("bPOCrefbias.csv")
np.savetxt('bPOCrefbias.csv',[maxbias])

os.chdir(dir+'bbp700')
bbp700stats.to_csv('bbp700'+gen)
os.chdir(dir+'chla')
chlastats.to_csv('chla'+gen)

END
)"
}

makecsv $MODEL


### 2. If not the last generation, performs a genetic algorithm (Biased Random Key Genetic Algorithm
###    (http://mauricio.resende.info/doc/brkgaAPI.pdf) the set of parameters that are generated are saved onto 0000.csv,
###    to be used in the next generation.
###    If it is the last generation, process the outputs and save

PARAMETERS=parameters/
cd $MODEL$PARAMETERS

module load python

function GA {

GA="$(python - $1 $2 << END

import numpy as np
import random
import pandas
import os

def geneticalgorithm(input,vars):

# Biased Random Key Genetic Algorithm (BRKGA)
# One generation crossover/mutation
# input - dataframe ((P+1)*M), extra column for values of the cost function
# output - dataframe (P*M) of new parameters

# P = number of parameters
# N = number of chunks

    #(N,P) = input.shape
    (N,P) = np.shape(input)
    P=P-1

    E=0.25  # proportion of elite
    M=0.25  # proportion of mutants
    rho=0.7 # probability of each crossover being assigned to the elite chunk

    NE=int(N*E)+1 # number of elite chunks (rounded up)
    NM=int(N*M)+1 # number of mutant chunks (rounded up)
    NC=N-(NE+NM)  # number of crossover chunk

    # sorting the input by cost and selecting elite and non-elite chunks
    input=pandas.DataFrame(input)
    input=input.sort_values(input.shape[1]-1).values
    inputE=input[:NE,:]
    inputnE=input[NE:,:]

    # putting elite chunks into output
    output=input[:NE,:-1]

    # generating mutants and putting into output

    vars=pandas.read_csv(vars)

    for i in range(NM):
        r=np.zeros((1,P))
        for j in range(P):
            #r[0,j]=1.2*0.5*(2*random.random()-1)*(np.amax(input[:,j])-np.amin(input[:,j]))+((np.amax(input[:,j])+np.amin(input[:,j]))*0.5)
            ## when the first number is 1, this assigns random values within the range of min and max. increasing it will extend the range about the mid
            #r[0,j]=random.random()*(np.amax(input[:,j])-np.amin(input[:,j]))+np.amin(input[:,j])
            r[0,j]=random.random()*(vars['max'][j]-vars['min'][j])+vars['min'][j]
        output=np.append(output,r,axis=0)

    # crossover
    for i in range(NC):
        r=np.zeros((1,P))
        en=inputE[np.random.randint(inputE.shape[0], size=1), :]
        nen=inputnE[np.random.randint(inputnE.shape[0], size=1), :]
        for j in range(P):
            a=random.random()
            if a>rho:
                #r[0,j]=nen[0,j]
                r[0,j]=np.random.normal(nen[0,j],(np.amax(input[:,j])-np.amin(input[:,j]))*0.01)
                while r[0,j]<0.:
                    r[0,j]=np.random.normal(nen[0,j],(np.amax(input[:,j])-np.amin(input[:,j]))*0.01)
            else:
                #r[0,j]=en[0,j]
                r[0,j]=np.random.normal(en[0,j],(np.amax(input[:,j])-np.amin(input[:,j]))*0.01)
                while r[0,j]<0.:
                    r[0,j]=np.random.normal(en[0,j],(np.amax(input[:,j])-np.amin(input[:,j]))*0.01)
        output=np.append(output,r,axis=0)

    return output

os.chdir('$1')
gen=len(os.listdir('$1'))-1
gen="{0:0=4d}".format(gen)+'.csv'

A=np.genfromtxt(gen)

vars='$2'
B=geneticalgorithm(A,vars)
if os.path.exists("0000.csv"):
    os.remove("0000.csv")
np.savetxt('0000.csv',B)

END
)"
}

GA ${MODEL}parameters $VARIABLES
